"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ts_simple_ast_1 = require("ts-simple-ast");
function moveNode(node, destFile) {
    // we copy all the imports from node.getSourceFile() to destFile and the organize imports so the moved declaration don't miss any import
    addImportsToDestFile(node, destFile);
    // For each sourceFile that reference node, we add an import declaration to node but specifying nodeFile. (we "move" the oriinal import declarations only changing the specifier to point to destFile).
    const referencedSourceFiles = getReferences(node).map(r => r.getSourceFile()).filter((f, i, a) => a.indexOf(f) === i);
    referencedSourceFiles.forEach(f => {
        const newImports = f.getImportDeclarations().filter(i => i.getModuleSpecifierSourceFile() === node.getSourceFile()).map(i => (Object.assign({}, i.getStructure(), { moduleSpecifier: f.getRelativePathAsModuleSpecifierTo(destFile) })));
        f.addImportDeclarations(newImports);
        f.getImportDeclarations().filter(i => i.getModuleSpecifierSourceFile() === node.getSourceFile()).forEach(i => i.remove());
    });
    // move the declaration
    let finalNode;
    if (ts_simple_ast_1.TypeGuards.isClassDeclaration(node)) {
        finalNode = destFile.addClass(node.getStructure());
    }
    else if (ts_simple_ast_1.TypeGuards.isInterfaceDeclaration(node)) {
        finalNode = destFile.addInterface(node.getStructure());
    }
    else if (ts_simple_ast_1.TypeGuards.isFunctionDeclaration(node)) {
        finalNode = destFile.addFunction(node.getStructure());
    }
    node.remove();
    finalNode.setIsExported(true);
    destFile.organizeImports();
    node.getSourceFile().organizeImports();
}
exports.moveNode = moveNode;
function getReferences(node) {
    const references = [];
    const referencedSymbols = node.findReferences();
    for (const referencedSymbol of referencedSymbols) {
        for (const reference of referencedSymbol.getReferences()) {
            references.push(reference);
        }
    }
    return references;
}
const typescript_ast_util_1 = require("typescript-ast-util");
function findReferencesDeclaredOutside(node, outside = true) {
    const refs = node.getDescendants().filter(ts_simple_ast_1.TypeGuards.isReferenceFindableNode)
        .map(d => {
        return getReferences(d)
            .filter(r => r.isDefinition() && (outside ? !r.getNode().getFirstAncestor(a => a === node) : !!r.getNode().getFirstAncestor(a => a === node))).map(r => r.getNode());
    });
    return typescript_ast_util_1.flat(refs).filter((n, i, a) => a.indexOf(n) === i);
}
exports.findReferencesDeclaredOutside = findReferencesDeclaredOutside;
function addImportsToDestFile(node, destFile) {
    const importsToAddToDestFile = node.getSourceFile().getImportDeclarations().filter(i => i.getModuleSpecifierSourceFile() !== destFile).map(i => {
        return Object.assign({}, i.getStructure(), { moduleSpecifier: destFile.getRelativePathAsModuleSpecifierTo(i.getModuleSpecifierSourceFile()) });
    });
    // import all nodes used by node that are declared outside it but not imported . Make sure they are exported
    const referencedByNodeNotImported = findReferencesDeclaredOutside(node)
        .filter(r => r.getSourceFile() === node.getSourceFile() && !r.getFirstAncestor(ts_simple_ast_1.TypeGuards.isImportDeclaration))
        .map(n => {
        const exportableAncestor = n.getFirstAncestor(a => ts_simple_ast_1.TypeGuards.isExportableNode(a) && ts_simple_ast_1.TypeGuards.isNameableNode(a) && a.getName() === n.getText());
        if (exportableAncestor) {
            exportableAncestor.setIsExported(true);
        }
        return n;
    })
        .filter((n, i, a) => a.indexOf(n) === i);
    referencedByNodeNotImported.forEach(n => {
        importsToAddToDestFile.push({
            moduleSpecifier: destFile.getRelativePathAsModuleSpecifierTo(node.getSourceFile()),
            namedImports: [{ name: n.getText() }]
        });
    });
    destFile.addImportDeclarations(importsToAddToDestFile);
}
exports.addImportsToDestFile = addImportsToDestFile;
//# sourceMappingURL=moveNode.js.map